#include "duplink.h"
#include <malloc.h>	// _alloca

HANDLE treeheap;
FileTree*root,**list;
FileTree::Count FileTree::c;

FileTree*FileTree::newItem(const TCHAR*parentpath,const WIN32_FIND_DATA*fd) {
 int l1 = lstrlen(parentpath);	// muss mit T('\\') enden, auer fr root
 int l2 = fd ? lstrlen(fd->cFileName) : 0;
 bool isDir = !fd || fd->dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY;
 bool follow = isDir && !(fd && fd->dwFileAttributes&FILE_ATTRIBUTE_REPARSE_POINT);
 FileTree*item = (FileTree*)HeapAlloc(treeheap,HEAP_ZERO_MEMORY|HEAP_NO_SERIALIZE,sizeof(FileTree)+(l1+l2+isDir+1)*sizeof(TCHAR));
 if (item) {
  item->iName = l1;
  memcpy(item->path,parentpath,(l1+1)*sizeof(TCHAR));
  if (fd) {
   memcpy(&item->Attr,&fd->dwFileAttributes,4+8+8+8);
   ((DWORD*)&item->Size)[0] = fd->nFileSizeLow;
   ((DWORD*)&item->Size)[1] = fd->nFileSizeHigh;
   memcpy(item->path+l1,fd->cFileName,(l2+1)*sizeof(TCHAR));
   if (isDir) {
    if (item->Attr&FILE_ATTRIBUTE_REPARSE_POINT) ++c.links;
    ++c.dirs;
    memcpy(item->path+l1+l2,T("\\"),2*sizeof(TCHAR));
   }else{
    if (item->Attr&FILE_ATTRIBUTE_REPARSE_POINT) ++c.links;
    else{
     ++c.files;
     c.bytes+=item->Size;
    }
   }
  }
  if (follow) item->sub = newTree(item);
 }else DebugBreak();	// Kein Heap-Speicher
 return item;
}

FileTree*FileTree::newTree(const FileTree*p) {
 int l1=lstrlen(p->path);	// muss ein Verzeichnis mit "\" am Ende sein!
 TCHAR*s = (TCHAR*)_alloca((l1+2)*sizeof(TCHAR));
 memcpy(s,p->path,l1*sizeof(TCHAR));
 memcpy(s+l1,T("*"),2*sizeof(TCHAR));
 WIN32_FIND_DATA fd;
 HANDLE h=FindFirstFile(s,&fd);
 FileTree*cur=0,*fst=0;
 if (h!=INVALID_HANDLE_VALUE) {
  do{
   if (!StrCmp(fd.cFileName,T("."))) continue;
   if (!StrCmp(fd.cFileName,T(".."))) continue;
   FileTree*nxt = newItem(p->path,&fd);	// fllt bereits nxt->sub
   if (!nxt) break;
   nxt->parent = p;
   nxt->prev = cur;
   if (cur) cur->next = nxt; else fst = nxt;
   cur = nxt;
  }while (FindNextFile(h,&fd));
  if (!FindClose(h)) DebugBreak();
 }
 return fst;
}

const FileTree**FileTree::flattenFiles(const FileTree**tgt) const{
 if (tgt) {	// Rekursionsbetrieb
  const FileTree*cur = sub;
  while (cur) {
   if (!(cur->Attr&FILE_ATTRIBUTE_REPARSE_POINT)) {	// Alle Arten von Links bergehen
    if (cur->Attr&FILE_ATTRIBUTE_DIRECTORY) tgt = cur->flattenFiles(tgt);
    else if (cur->Size) *tgt++ = cur;	// Null-Byte-Dateien ignorieren
   }
   cur = cur->next;
  }
  return tgt;	// Neuer Fllstand
 }
 const FileTree**lst = (const FileTree**)HeapAlloc(treeheap,HEAP_ZERO_MEMORY|HEAP_NO_SERIALIZE,c.files*sizeof(FileTree*));
 if (lst) {	// noch Platz im Heap
  const FileTree**tgt = flattenFiles(lst);
  if (unsigned(tgt-lst) > c.files) DebugBreak();	// Problemfall: Heap korrumpiert
  if (c.files > unsigned(tgt-lst)) {	// Links und 0-Byte-Dateien nicht in Liste
   c.files = tgt-lst;
   HeapReAlloc(treeheap,HEAP_NO_SERIALIZE,lst,c.files*sizeof(FileTree*));	// Liste kleiner machen
  }
 }
 return lst;
}

int FileTree::depth() const {	// Backslashs zhlen
 const TCHAR *q, *p = path;
 int ret = 0;
 while (q=StrChr(p,'\\')) {
  ++ret;
  p = q+1;
 }
 return ret;
}

int FileTree::hiddenDirs() const {
 const TCHAR *q, *p = path;
 int ret = 0;
 while (q=StrChr(p,'\\')) {
  if (*p=='.') ++ret;
  p = q+1;
 }
 return ret;
}


// C++20: operator<=>
int FileTree::comparefunc(const FileTree*other) const {
// Garantiert Dateien, die keine Links ("Reparse Points") sind.
// Nach Gre sortieren
 if (Size<other->Size) return -1;
 if (Size>other->Size) return +1;
// Nach Dateiname sortieren
 int i=StrCmp(name(),other->name());
 if (i) return i;
// Nach Verstecktheit von Verzeichnissen sortieren (nur fr mich)
 i = hiddenDirs() - other->hiddenDirs();
 if (i) return i;
// Nach Pfadlnge sortieren (Anzahl der Backslashs)
 i = depth() - other->depth();
 if (i) return i;
// Nach Pfadlnge sortieren (kurze Pfade bevorzugen)
 i = iName - other->iName;
 if (i) return i;
// Nach Pfad alphabetisch sortieren
 return StrCmp(path,other->path);
}


